// $Id: X11Interface.cpp,v 1.6 2014/01/15 21:55:02 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "X11Interface.h"

#include "Debug.h"
#include "X11Util.h"
#include "X11ErrorHandler.h"

#if defined(Q_WS_X11) || defined( Q5_WS_X11 )
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#endif

//_________________________________________________________
void X11Interface::findIcons( Job::List& jobs ) const
{

    // check atoms
    if( !( X11Util::get().isSupported( X11Util::_NET_WM_PID ) && X11Util::get().isSupported( X11Util::_NET_WM_ICON ) ) )
    { return; }

    // find recursively, starting from root window
    _findIcons( X11Util::get().appRootWindow(), jobs );

}

//_________________________________________________________
void X11Interface::_findIcons( WId window, Job::List& jobs ) const
{

    #if defined(Q_WS_X11) || defined( Q5_WS_X11 )
    Debug::Throw() << "X11Interface::_findIcons - window: " << window << endl;

    // get display and atom
    Display* display( reinterpret_cast<Display*>( X11Util::get().display() ) );
    Atom pidAtom( reinterpret_cast<Atom>( X11Util::get().findAtom( X11Util::_NET_WM_PID ) ) );

    // install error handler
    X11ErrorHandler errorHandler;
    Q_UNUSED( errorHandler );

    Atom type(None);
    int format(0);
    unsigned long count(0);
    unsigned long after(0);
    unsigned char *property(0);

    // get pid
    if( XGetWindowProperty( display, window, pidAtom, 0, 1, False, XA_CARDINAL, &type, &format, &count, &after, &property ) == Success && property )
    {

        // If the PID matches, add this window to the result set.
        unsigned long pid = *(reinterpret_cast<unsigned long*>(property) );
        XFree(property);

        // find matching job in list
        Job::List::iterator jobIter = std::find_if( jobs.begin(), jobs.end(), Job::SameIdFTor( pid ) );
        if( jobIter != jobs.end() )
        {
            // check job pixmap status
            // no need to go further, nor find children windows if jobIter already has a pixmap
            if( jobIter->iconInitialized() ) return;

            // find window icon
            QIcon icon( _icon( window ) );
            if( !icon.isNull() )
            {
                jobIter->setIconInitialized( true );
                jobIter->setIcon( icon );
                return;
            }
        }
    }

    // parse children
    Window root = None;
    Window parent = None;
    Window* children = 0;
    unsigned int nChildren( 0 );
    if( XQueryTree( display, window, &root, &parent, &children, &nChildren ) )
    {
        for(unsigned i = 0; i < nChildren; i++)
        { _findIcons( children[i], jobs ); }
    }

    #endif
    return;

}

//_________________________________________________________
QIcon X11Interface::_icon( WId window ) const
{

    QIcon result;

    #if defined(Q_WS_X11) || defined( Q5_WS_X11 )
    Debug::Throw() << "X11Interface::_icon - window: " << window << endl;

    // install error handler
    X11ErrorHandler errorHandler;
    Q_UNUSED( errorHandler );

    // get display and atom
    Display* display( reinterpret_cast<Display*>( X11Util::get().display() ) );
    Atom iconAtom( reinterpret_cast<Atom>( X11Util::get().findAtom( X11Util::_NET_WM_ICON ) ) );

    Atom type = None;
    int format = 0;
    unsigned long count(0);
    unsigned long after(0);
    unsigned char *property(0);
    int offset(0);
    while( true )
    {
        // get width
        int width(0);
        if( XGetWindowProperty( display, window, iconAtom, offset, 1, 0, XA_CARDINAL, &type, &format, &count, &after, &property ) == Success && property )
        {

            width = *(reinterpret_cast<int*>(property) );
            XFree( property );

        } else break;

        // get height
        int height(0);
        property = 0;
        if( XGetWindowProperty( display, window, iconAtom, offset+1, 1, 0, XA_CARDINAL, &type, &format, &count, &after, &property ) == Success && property )
        {

            height = *(reinterpret_cast<int*>(property) );
            XFree( property );

        } else break;

        // pixmap
        ulong* pixmapData;
        const int size( width*height );
        if( XGetWindowProperty( display, window, iconAtom, offset+2, size, 0, XA_CARDINAL, &type, &format, &count, &after, (uchar**)&pixmapData ) == Success )
        {
            QImage image( width, height, QImage::Format_ARGB32);
            for(int i=0; i<image.byteCount()/4; ++i)
            { ((uint*)image.bits())[i] = pixmapData[i]; }
            result.addPixmap( QPixmap::fromImage( image ) );
            XFree( pixmapData );
        }

        if( !after ) break;
        offset += (2+width*height);

    }

    #endif

    return result;
}
